/**
 * encapsulates all the steps for a default crud panel
 * 1- get column information
 * 2- get the actual data (from cache if available)
 * 3- filter only the visible information
 *
 * @param {string} entityName the database entity's name
 * @param {Number} userId the requester id
 * @param {Connection} conn database connection
 * @returns an object that contains the data to be displayed and also the columns information to define how the table will be drawn
 */
module.exports.crudData = async function (entityName, userId, conn) {
  if (!conn) conn = await this.newConnectionFromPool(false);
  const upperEntityName = stringUtils.underlineToUppercase(entityName, true); // the key for the values on the database

  const configKey = `pagination${upperEntityName}`; // the key for the result values on the database

  const cacheDataKey = `data${upperEntityName}`;
  const columns = await this.getColumInfo(entityName, userId, conn);
  let data = await this.execWithCache({
    sql: `select * from \`${configKey}\``
  }, conn, cacheDataKey, configKey);

  if (data.length > 0) {
    const visibleColumns = columns.filter(i => i.field === `id` || i.field.endsWith(`Id`) || !i.hidden).map(i => i.field);
    data = arrayUtils.getPropertySet(data, visibleColumns);
  }

  return {
    data,
    columns
  };
};
/**
 * inserts a list of items on a given table
 *
 * @param {Connection} conn database connection
 * @param {Function} action the action to be taken
 * @param {number} userId the id of the user who initiated the action
 * @param {any[]} params the action parameters
 * @returns {{insertedItems: any}} the items who were just inserted (with their respective new ids)
 */


module.exports.insertItems = async function (conn, action, userId, params) {
  try {
    const returnData = [];
    if (typeUtils.isArrayAndValid(params)) for (let param of params) returnData.push(await action(conn, userId, param));else returnData.push(await action(conn, userId, params));
    return {
      insertedItems: returnData
    };
  } catch (err) {
    throw {
      code: err.errno || 0,
      devMessage: err.sqlMessage || err.message,
      humanMessage: `Erro ao inserir um novo(s) item(s)`
    };
  }
};
/**
 * deletes an item, or items, from a table
 *
 * @param {string} table the table to be affected
 * @param {number[]} itemIds the id's to be affected
 * @param {number} userId the user's id
 * @param {Connection} conn database connectio
 * @returns {{success: boolean, changedRows: number}}a result object with a boolean (success) and an integer representing the number of affected rows
 */


module.exports.deleteItems = async function (table, itemIds, userId, conn) {
  if (!conn) conn = await this.newConnectionFromPool(false);
  const questionMarks = itemIds.map(i => `?`);
  const queryOptions = {
    sql: `update \`${table}\` set \`removed\` = 1 where \`id\` in (${questionMarks})`,
    values: itemIds
  };
  const result = await this.exec(queryOptions, conn);
  if (!result) return {
    success: false,
    changedRows: null
  };else return {
    success: true,
    changedRows: result.changedRows
  };
};

const crudConfigUser = `configUser`;
const crudConfigTable = `configCrud`;
/**
 * gets the column information for a crud view
 *
 * @param {string} entityName the database entity's name
 * @param {Number} userId the userId, if provided this will indecate a user level crud customization
 * @param {Connection} conn database connection
 */

module.exports.getColumInfo = async (entityName, userId, conn) => !userId ? getTableColumns(entityName, conn) : getUserColumns(entityName, userId, conn);

const getUserColumns = async (entityName, userId, conn) => {
  const upperEntityName = stringUtils.firstLetterUppercase(entityName); // the key for the values on the database

  const configKey = `pagination${upperEntityName}`; // the key for the values on the internal cache

  const cacheColumnsKey = `columns${upperEntityName}`;
  const userEntry = `${cacheColumnsKey}${userId}`;
  if (mysqlCache.has(userEntry)) return mysqlCache.load(userEntry);else {
    const queryColumns = {
      sql: `select \`params\` from \`${crudConfigUser}\` where \`configuration\` = ? and \`userId\` = ?`,
      values: [configKey, userId]
    };
    const columnsData = await this.exec(queryColumns, conn);

    if (typeUtils.isArrayAndValid(columnsData)) {
      const result = JSON.parse(columnsData[0].params);
      mysqlCache.store(userEntry, result);
      return result;
    } else return await getTableColumns(entityName, conn);
  }
};

const getTableColumns = async (entityName, conn) => {
  const upperEntityName = stringUtils.firstLetterUppercase(entityName);
  const configKey = `pagination${upperEntityName}`;
  const cacheColumnsKey = `columns${upperEntityName}`;
  if (mysqlCache.has(cacheColumnsKey)) return mysqlCache.load(cacheColumnsKey);else {
    const queryColumns = {
      sql: `select \`params\` from \`${crudConfigTable}\` where \`configuration\` = ?`,
      values: [configKey]
    };
    const columnsData = await this.exec(queryColumns, conn);

    if (typeUtils.isArrayAndValid(columnsData)) {
      const result = JSON.parse(columnsData[0].params);
      mysqlCache.store(cacheColumnsKey, result);
      return result;
    } else throw new Error(`MysqlManager getTableColumns: there is no column info for entry '${configKey}'`);
  }
};
/**
 * stores a change in the display settings for a crudPanel
 *
 * @param {string} entityName the entity's name
 * @param {Number} userId the requester id
 * @param {string} newDefinition the new definition
 * @param {Connection} conn database connection
 */


module.exports.setColumInfo = async (entityName, userId, newDefinition, conn) => {
  const upperEntityName = stringUtils.firstLetterUppercase(entityName);
  const configKey = `pagination${upperEntityName}`;
  const cacheColumnsKey = `columns${upperEntityName}`;
  const userEntry = `${cacheColumnsKey}${userId}`;
  const columnsInfo = await this.getColumInfo(entityName, userId, conn);

  for (let columnInfo of columnsInfo) columnInfo.hidden = newDefinition[columnInfo.field];

  const columnValues = JSON.stringify(columnsInfo);
  let queryOptions = {
    sql: ``,
    values: []
  };

  if (userId) {
    queryOptions.sql = `update \`${crudConfigUser}\` set \`params\` = ? where \`configuration\` = ? and \`userId\` = ?`;
    queryOptions.values = [columnValues, configKey, userId];
  } else {
    queryOptions.sql = `update \`${crudConfigTable}\` set \`params\` = ? where \`configuration\` = ?`;
    queryOptions.values = [columnValues, configKey];
  }

  mysqlCache.store(userId ? userEntry : configKey, columnsInfo);
  const result = await this.exec(queryOptions, conn);
  if (!result) return {
    success: false,
    changedRows: null
  };else return {
    success: true,
    changedRows: result.changedRows
  };
};