const typeUtils = require('clericuzzi-javascript/dist/utils/type.utils');

const stringUtils = require('clericuzzi-javascript/dist/utils/string.utils');

const propertyNameUtils = require('../utils/propertyName.utils');

module.exports = class CrudBase {
  /**
  * Creates an instance of CrudBase.
  * @param {string} tableName the represented table's name
  * @param {[string]} columns the columns list
  * @param {[string]} dataTypes the data types list
  * @param {[string]} nullableColumns the nullable columns list
  * @param {[string]} fkColumns thr fks list
  * @param {boolean} hasRemoved if TRUE the table has the 'removed' field
  * @param {string} idProperty the id property's name
  * @param {any} schema the protobuf schema
  * @param {string} databaseType the kind of database engine
  */
  constructor(tableName, columns, dataTypes, nullableColumns, fkColumns, hasRemoved, idProperty, schema, databaseType) {
    if (hasRemoved) this.removed = 0;
    this.tableName = tableName;
    this.idProperty = idProperty;
    this.hasRemoved = hasRemoved;
    this.databaseType = databaseType;
    this.columns = columns;
    this.fkColumns = fkColumns;
    this.nullableColumns = nullableColumns;
    this.dataTypes = dataTypes;
    this.schema = schema;
  }
  /**
   * copies the object's values for itself
   *
   * @param {Object} data the params object that contains the model data
   */


  loadValues(data) {
    for (let column of this.columns) this[column] = typeUtils.isValid(data[column]) ? data[column] : typeUtils.isValid(this[column]) ? this[column] : null;
  }

  toJson(stringify = true) {
    const obj = {};

    for (let column of this.columns) obj[column] = this[column];

    for (let column of this.columns) {
      if (this.fkColumns && this.fkColumns.includes(column)) {
        obj[column] = this[column] && this[column].toJson ? this[column].toJson(false) : this[column];
        const fkProperty = column.slice(0, -2);
        if (this[fkProperty]) obj[fkProperty] = this[fkProperty] && this[fkProperty].toJson ? this[fkProperty].toJson(false) : this[fkProperty];
      }
    }

    if (stringify) return JSON.stringify(obj);else return obj;
  }

  fromJson(json) {
    const jsonKeys = Object.keys(json);

    for (let jsonKey of jsonKeys) if (this.columns.includes(jsonKey)) this[jsonKey] = typeUtils.isValid(json[jsonKey]) ? json[jsonKey] : null;
  }
  /**
   * create an instance of the pb schema object
   *
   * @returns {*} the schema object instance
   */


  toSchema() {
    if (!this.schema) throw new Error(`Protobuf not defined for this item (${this.tableName})`);
    const setNames = this.columns.map(i => `set` + stringUtils.firstLetterUppercase(i.toLocaleLowerCase()));

    for (let fkColumn of this.fkColumns) setNames.push(`set${stringUtils.firstLetterUppercase(propertyNameUtils.getForeignPropertyName(fkColumn).toLocaleLowerCase())}`);

    const pb = new this.schema();
    let index = 0;

    for (let setName of setNames) {
      if (index === this.columns.length) break;
      pb[setName](this[this.columns[index++]]);
    }

    for (let fkColumn of this.fkColumns) {
      const methodName = setNames[index++];

      if (pb[methodName]) {
        const propertyName = propertyNameUtils.getForeignPropertyName(fkColumn);
        let propertyValue = this[propertyName];
        if (propertyValue === undefined) propertyValue = null;else propertyValue = propertyValue.toSchema();
        pb[methodName](propertyValue);
      }
    }

    return pb;
  }
  /**
   * restores this instance based in a pb schema
   *
   * @param {*} schema the schema representation
   */


  fromSchema(pb) {
    if (!this.schema) throw new Error(`Protobuf not defined for this item (${this.tableName})`);
    const getNames = this.columns.map(i => `get` + stringUtils.firstLetterUppercase(i.toLocaleLowerCase()));

    for (let fkColumn of this.fkColumns) getNames.push(`get${stringUtils.firstLetterUppercase(propertyNameUtils.getForeignPropertyName(fkColumn).toLocaleLowerCase())}`);

    let index = 0;

    for (let getName of getNames) {
      if (index === this.columns.length) break;
      this[this.columns[index++]] = pb[getName]();
    }

    for (let fkColumn of this.fkColumns) if (this[propertyNameUtils.getForeignPropertyName(fkColumn)]) {
      this[propertyNameUtils.getForeignPropertyName(fkColumn)] = pb[getNames[index++]]();
      if (this[propertyNameUtils.getForeignPropertyName(fkColumn)] === undefined) this[propertyNameUtils.getForeignPropertyName(fkColumn)] = null;
    }

    return this;
  }
  /**
   * writes a byte array representation of the current instance
   *
   * @returns {[byte]} the byte array
   */


  toBinary() {
    return this.toSchema().serializeBinary();
  }
  /**
   * restores this instance based in a byte array representation
   *
   * @param {[byte]} bytes the byte array representation
   */


  fromBinary(bytes) {
    return this.fromSchema(this.schema.deserializeBinary(bytes));
  }
  /**
   * validates the entity structure
   */


  validate() {
    if (!this.tableName) throw {
      code: 0,
      devMessage: `${this.databaseType} CrudModel: '${this.tableName}' table not initialized`,
      humanMessage: `Falha ao iniciar a tabela '${this.tableName}'`
    };
    if (this.columns.size === 0) throw {
      code: 0,
      devMessage: `${this.databaseType} CrudModel: '${this.tableName}' columns not initialized`,
      humanMessage: `Falha ao iniciar as colunas da tabela '${this.tableName}'`
    };
  }
  /**
   * validates a given priperty
   *
   * @param {string} propertyName the property to be validated
   */


  validateProperty(propertyName) {
    if (this.columns.size === 0) throw {
      code: 0,
      devMessage: `${this.databaseType} CrudModel: '${this.tableName}' columns not initialized`,
      humanMessage: `Falha ao iniciar as colunas da tabela '${this.tableName}'`
    };
    if (!this.columns.includes(propertyName)) throw {
      code: 0,
      devMessage: `${this.databaseType} CrudModel: '${this.tableName}' columns not initialized`,
      humanMessage: `Falha ao iniciar as coluna '${propertyName}' da tabela '${this.tableName}'`
    };
    const columnIndex = this.columns.indexOf(propertyName);
    const isNullable = this.nullableColumns[columnIndex];
    const value = this[propertyName];
    if (isNullable && (value === undefined || value === null || value === NaN)) throw {
      code: 0,
      devMessage: `${this.databaseType} CrudModel: '${this.tableName}' not filled correctly`,
      humanMessage: `Falha ao preencher as coluna '${propertyName}' da tabela '${this.tableName}'`
    };
  }

  loadProperty(property, value) {
    if (this.fkColumns && this.fkColumns.includes(property)) {
      const foreignProperty = property.slice(0, -2);
      if (this[foreignProperty]) this[foreignProperty][`id`] = value;
    }

    this[property] = value;
  }

};